/* This file is part of scl, a control and simulation library
for robots and biomechanical models.

scl is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

Alternatively, you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

scl is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License and a copy of the GNU General Public License along with
scl. If not, see <http://www.gnu.org/licenses/>.
 */
/* \file scl_tutorial4_control_gc_op.cpp
 *
 *  Created on: Jul 30, 2014
 *
 *  Copyright (C) 2014
 *
 *  Author: Samir Menon <smenon@stanford.edu>
 */

//scl lib

#include <scl/DataTypes.hpp>
#include <scl/data_structs/SGcModel.hpp>
#include <scl/dynamics/scl/CDynamicsScl.hpp>
#include <scl/dynamics/tao/CDynamicsTao.hpp>
#include <scl/parser/sclparser/CParserScl.hpp>
#include <scl/graphics/chai/CGraphicsChai.hpp>
#include <scl/graphics/chai/ChaiGlutHandlers.hpp>
#include <scl/util/DatabaseUtils.hpp>

//global database of everything
#include <scl/Singletons.hpp>
#include <scl/util/DatabaseUtils.hpp>
#include <scl/util/HelperFunctions.hpp>

//For timing
#include <sutil/CSystemClock.hpp>

//Eigen 3rd party lib
#include <Eigen/Dense>

//Standard includes (for printing and multi-threading)
#include <iostream>
#include <omp.h>

//Freeglut windowing environment
#include <GL/freeglut.h>

/** A sample application to demonstrate a physics simulation in scl.
 *
 * Moving forward from tutorial 3, we will now control the 6 DOF
 * demo robot (r6bot) with the physics engine running.
 *
 * SCL Modules used:
 * 1. data_structs
 * 2. dynamics (physics)
 * 4. dynamics (control matrices)
 * 4. graphics (chai)
 * */
int main(int argc, char** argv)
{
    scl::CParserScl parser;
    scl::SRobotParsed robot_properties;

    scl_util::getCurrentDir(scl::CDatabase::getData()->cwd_);
    scl::CDatabase::getData()->dir_specs_ = scl::CDatabase::getData()->cwd_;

    //parsed the robot
    parser.readRobotFromFile("./Puma/PumaCfg.xml", "PumaBot", robot_properties);

    //sensors, input/outputs
    scl::SRobotIO robotIO;
    robotIO.init(robot_properties.name_, robot_properties.dof_);

    //set initial sensor values
    Eigen::VectorXd q_default(robot_properties.dof_);

    for ( unsigned int i=0; i<robot_properties.dof_; ++i) {
        scl::SRigidBody* rigid_body = robot_properties.rb_tree_.at(i);
        q_default(i) = rigid_body->joint_default_pos_;
    }

    robotIO.sensors_.q_ = q_default;

    //create physics engine
    scl::CDynamicsTao world_dynamics;
    world_dynamics.init(robot_properties);


    //dynamic robot
    scl::SGcModel robot_model;
    robot_model.init(robot_properties);
    scl::SRigidBodyDyn *hand = robot_model.rbdyn_tree_.at("end-effector");
    const Eigen::Vector3d pos_point_in_hand(0,0,0);


    //our control dynamics engine
    scl::CDynamicsScl robot_dynamics;
    robot_dynamics.init(robot_properties);



    //graphics
    glutInit(&argc, argv);

    scl::SGraphicsParsed graphics_properties;
    parser.readGraphicsFromFile("./Puma/PumaCfg.xml", "PumaBotStdView", graphics_properties);

    scl::CGraphicsChai graphics;
    graphics.initGraphics(&graphics_properties);
    graphics.addRobotToRender(&robot_properties, &robotIO);
    scl_chai_glut_interface::initializeGlutForChai(&graphics_properties, &graphics);

    //create two threads
    omp_set_num_threads(2);
    int thread_id;

    //create timer for integrating
    double tlast, tcurr, dt;
    sutil::CSystemClock::start();


#pragma omp parallel private(thread_id)
    {
        thread_id = omp_get_thread_num();

        if (thread_id==1) { //simulation for physics and control

            tlast = sutil::CSystemClock::getSysTime();

            while(true) {
                //step the time
                tcurr = sutil::CSystemClock::getSysTime();
                dt = tcurr - tlast;

                //get hand position
                robot_dynamics.computeTransformsForAllLinks(robot_model.rbdyn_tree_,
                                                            robotIO.sensors_.q_);
                Eigen::Vector3d pos = hand->T_o_lnk_*pos_point_in_hand;

                std::cout << pos << std::endl;

                //PD joint space controller
                int Kp = 100;
                int Kd = 20;

                robotIO.actuators_.force_gc_commanded_ = Kp*(q_default - robotIO.sensors_.q_)
                                                        -Kd*(robotIO.sensors_.dq_);


                //integrate dynamics
                world_dynamics.integrate(robotIO, dt);

                const timespec ts = {0,5000};
                nanosleep(&ts, NULL);

                tlast=tcurr;
            }
        }

        else { //thread 2 for graphics
            while(true) {
                glutMainLoop();
                const timespec ts = {0,15000000};
                nanosleep(&ts, NULL);

            }
        }
    }

  return 0;
}



















